package com.mapmyindia.mapmyindiagl.api;

import androidx.annotation.NonNull;

import com.google.gson.Gson;
import com.mapmyindia.sdk.geoanalytics.listing.MapmyIndiaGeoAnalyticsListManager;
import com.mapmyindia.sdk.geoanalytics.listing.model.GeoAnalyticsListResponse;
import com.mmi.services.api.OnResponseCallback;
import com.mmi.services.api.PlaceResponse;
import com.mmi.services.api.alongroute.MapmyIndiaPOIAlongRouteManager;
import com.mmi.services.api.alongroute.models.POIAlongRouteResponse;
import com.mmi.services.api.autosuggest.MapmyIndiaAutosuggestManager;
import com.mmi.services.api.autosuggest.model.AutoSuggestAtlasResponse;
import com.mmi.services.api.directions.MapmyIndiaDirectionManager;
import com.mmi.services.api.directions.models.DirectionsResponse;
import com.mmi.services.api.distance.MapmyIndiaDistanceMatrixManager;
import com.mmi.services.api.distance.models.DistanceResponse;
import com.mmi.services.api.eloc.MapmyIndiaELocManager;
import com.mmi.services.api.event.nearby.MapmyIndiaNearbyReportManager;
import com.mmi.services.api.event.nearby.model.NearbyReportResponse;
import com.mmi.services.api.feedback.MapmyIndiaFeedbackManager;
import com.mmi.services.api.generateotp.MapmyIndiaGenerateOTP;
import com.mmi.services.api.generateotp.MapmyIndiaGenerateOtpManager;
import com.mmi.services.api.geocoding.GeoCodeResponse;
import com.mmi.services.api.geocoding.MapmyIndiaGeoCodingManager;
import com.mmi.services.api.hateaosnearby.MapmyIndiaHateosNearby;
import com.mmi.services.api.hateaosnearby.MapmyIndiaHateosNearbyManager;
import com.mmi.services.api.nearby.MapmyIndiaNearbyManager;
import com.mmi.services.api.nearby.model.NearbyAtlasResponse;
import com.mmi.services.api.placedetail.MapmyIndiaPlaceDetailManager;
import com.mmi.services.api.placedetail.model.PlaceDetailResponse;
import com.mmi.services.api.reversegeocode.MapmyIndiaReverseGeoCodeManager;
import com.mmi.services.api.textsearch.MapmyIndiaTextSearchManager;
import com.mmi.services.api.traffic.MapmyIndiaRoadTrafficDetailManager;
import com.mmi.services.api.traffic.model.TrafficRoadDetailResponse;
import com.mmi.services.api.whitelist.MapmyIndiaWhitelistManager;

import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.Map;

import io.flutter.plugin.common.MethodCall;
import io.flutter.plugin.common.MethodChannel;
import retrofit2.Call;
import retrofit2.Callback;
import retrofit2.Response;

/**
 * * Created by Saksham on 14-05-2021.
 **/
public class RestApiMethodHandler implements MethodChannel.MethodCallHandler {


    @Override
    public void onMethodCall(@NonNull MethodCall call, @NonNull MethodChannel.Result result) {
        switch (call.method) {
            case "restApi#autosuggest":
                MapmyIndiaAutosuggestManager.newInstance(RestApiHelper.autoSuggest(call.arguments))
                        .call(new OnResponseCallback<AutoSuggestAtlasResponse>() {
                            @Override
                            public void onSuccess(AutoSuggestAtlasResponse autoSuggestAtlasResponse) {
                                result.success(autoSuggestAtlasResponse != null ? new Gson().toJson(autoSuggestAtlasResponse) : null);
                            }

                            @Override
                            public void onError(int i, String s) {
                                result.error(i + "", s, null);
                            }
                        });
                break;

            case "rest#geocode":
                MapmyIndiaGeoCodingManager.newInstance(RestApiHelper.geoCoding(call.arguments))
                        .call(new OnResponseCallback<GeoCodeResponse>() {
                            @Override
                            public void onSuccess(GeoCodeResponse geoCodeResponse) {
                                result.success(geoCodeResponse != null ? new Gson().toJson(geoCodeResponse) : null);
                            }

                            @Override
                            public void onError(int i, String s) {
                                result.error(i + "", s, null);
                            }
                        });
                break;

            case "rest#revgeocode":
                MapmyIndiaReverseGeoCodeManager.newInstance(RestApiHelper.reverseGeoCode(call.arguments))
                        .call(new OnResponseCallback<PlaceResponse>() {
                            @Override
                            public void onSuccess(PlaceResponse placeResponse) {
                                result.success(placeResponse != null ? new Gson().toJson(placeResponse) : null);
                            }

                            @Override
                            public void onError(int i, String s) {
                                result.error(i + "", s, null);
                            }
                        });
                break;

            case "rest#nearby":
                MapmyIndiaNearbyManager.newInstance(RestApiHelper.nearby(call.arguments))
                        .call(new OnResponseCallback<NearbyAtlasResponse>() {
                            @Override
                            public void onSuccess(NearbyAtlasResponse nearbyAtlasResponse) {
                                result.success(nearbyAtlasResponse != null ? new Gson().toJson(nearbyAtlasResponse) : null);
                            }

                            @Override
                            public void onError(int i, String s) {
                                result.error(i + "", s, null);
                            }
                        });
                break;

            case "rest#direction":
                MapmyIndiaDirectionManager.newInstance(RestApiHelper.directions(call.arguments))
                        .call(new OnResponseCallback<DirectionsResponse>() {
                            @Override
                            public void onSuccess(DirectionsResponse directionsResponse) {
                                result.success(directionsResponse != null ? new Gson().toJson(directionsResponse) : null);
                            }

                            @Override
                            public void onError(int i, String s) {
                                result.error(i + "", s, null);
                            }
                        });
                break;

            case "rest#distancematrix":
                MapmyIndiaDistanceMatrixManager.newInstance(RestApiHelper.distanceMatrix(call.arguments))
                        .call(new OnResponseCallback<DistanceResponse>() {
                            @Override
                            public void onSuccess(DistanceResponse distanceResponse) {
                                result.success(distanceResponse != null ? new Gson().toJson(distanceResponse) : null);
                            }

                            @Override
                            public void onError(int i, String s) {
                                result.error(i + "", s, null);
                            }
                        });
                break;

            case "rest#feedback":
                MapmyIndiaFeedbackManager.newInstance(RestApiHelper.feedback(call.arguments))
                        .call(new OnResponseCallback<Void>() {
                            @Override
                            public void onSuccess(Void unused) {
                                result.success(201);
                            }

                            @Override
                            public void onError(int i, String s) {
                                result.error(i + "", s, null);
                            }
                        });
                break;

            case "rest#poialongroute":
                MapmyIndiaPOIAlongRouteManager.newInstance(RestApiHelper.poiAlongRoute(call.arguments))
                        .call(new OnResponseCallback<POIAlongRouteResponse>() {
                            @Override
                            public void onSuccess(POIAlongRouteResponse poiAlongRouteResponse) {
                                result.success(poiAlongRouteResponse != null ? new Gson().toJson(poiAlongRouteResponse) : null);
                            }

                            @Override
                            public void onError(int i, String s) {
                                result.error(i + "", s, null);
                            }
                        });
                break;

            case "rest#textsearch":
                MapmyIndiaTextSearchManager.newInstance(RestApiHelper.textSearch(call.arguments))
                        .call(new OnResponseCallback<AutoSuggestAtlasResponse>() {
                            @Override
                            public void onSuccess(AutoSuggestAtlasResponse autoSuggestAtlasResponse) {
                                result.success(autoSuggestAtlasResponse != null ? new Gson().toJson(autoSuggestAtlasResponse) : null);
                            }

                            @Override
                            public void onError(int i, String s) {
                                result.error(i + "", s, null);
                            }
                        });
                break;

            case "rest#placedetail":
                MapmyIndiaPlaceDetailManager.newInstance(RestApiHelper.placeDetail(call.arguments))
                        .call(new OnResponseCallback<PlaceDetailResponse>() {
                            @Override
                            public void onSuccess(PlaceDetailResponse placeDetailResponse) {
                                result.success(placeDetailResponse != null ? new Gson().toJson(placeDetailResponse) : null);
                            }

                            @Override
                            public void onError(int i, String s) {
                                result.error(i + "", s, null);
                            }
                        });
                break;

            case "rest#placedetaillegacy":
                MapmyIndiaELocManager.newInstance(RestApiHelper.placeDetailLegacy(call.arguments))
                        .call(new OnResponseCallback<PlaceResponse>() {
                            @Override
                            public void onSuccess(PlaceResponse placeResponse) {
                                result.success(placeResponse != null ? new Gson().toJson(placeResponse) : null);
                            }

                            @Override
                            public void onError(int i, String s) {
                                result.error(i + "", s, null);
                            }
                        });
                break;

            case "rest#hateosnearby":
                MapmyIndiaHateosNearbyManager.newInstance(RestApiHelper.hateOsNearby(call.arguments))
                        .call(new OnResponseCallback<NearbyAtlasResponse>() {
                            @Override
                            public void onSuccess(NearbyAtlasResponse nearbyAtlasResponse) {
                                result.success(nearbyAtlasResponse != null ? new Gson().toJson(nearbyAtlasResponse) : null);
                            }

                            @Override
                            public void onError(int i, String s) {
                                result.error(i + "", s, null);
                            }
                        });
                break;

            case "rest#geoanalyticslist":
                MapmyIndiaGeoAnalyticsListManager.newInstance(RestApiHelper.geoAnalyticsList(call.arguments))
                        .call(new OnResponseCallback<GeoAnalyticsListResponse>() {
                            @Override
                            public void onSuccess(GeoAnalyticsListResponse geoAnalyticsListResponse) {
                                result.success(geoAnalyticsListResponse != null ? new Gson().toJson(geoAnalyticsListResponse) : null);
                            }

                            @Override
                            public void onError(int i, String s) {
                                result.error(i + "", s, null);
                            }
                        });
                break;

            case "rest#generateotp":
                MapmyIndiaGenerateOtpManager.newInstance(RestApiHelper.generateOTP(call.arguments))
                        .call(new OnResponseCallback<String>() {
                            @Override
                            public void onSuccess(String s) {
                                result.success(s);
                            }

                            @Override
                            public void onError(int i, String s) {
                                result.error(i + "", s, null);
                            }
                        });
                break;
            case "rest#whitelist":
                MapmyIndiaWhitelistManager.newInstance(RestApiHelper.whitelist(call.arguments))
                        .call(new OnResponseCallback<Void>() {
                            @Override
                            public void onSuccess(Void unused) {
                                result.success(null);
                            }

                            @Override
                            public void onError(int i, String s) {
                                result.error(i + "", s, null);
                            }
                        });
                break;
            case "rest#roadtrafficdetail":
                MapmyIndiaRoadTrafficDetailManager.newInstance(RestApiHelper.roadTrafficDetail(call.arguments))
                        .call(new OnResponseCallback<TrafficRoadDetailResponse>() {
                            @Override
                            public void onSuccess(TrafficRoadDetailResponse trafficRoadDetailResponse) {
                                result.success(trafficRoadDetailResponse != null ? new Gson().toJson(trafficRoadDetailResponse) : null);
                            }

                            @Override
                            public void onError(int i, String s) {
                                result.error(i + "", s, null);
                            }
                        });
                break;
            case "rest#nearbyreport":
                MapmyIndiaNearbyReportManager.newInstance(RestApiHelper.nearbyReport(call.arguments))
                        .call(new OnResponseCallback<NearbyReportResponse>() {
                            @Override
                            public void onSuccess(NearbyReportResponse nearbyReportResponse) {
                                result.success(nearbyReportResponse != null ? new Gson().toJson(nearbyReportResponse) : null);
                            }

                            @Override
                            public void onError(int i, String s) {
                                result.error(i + "", s, null);
                            }
                        });
                break;
        }
    }

}
