import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:mapmyindia_gl_example/utils/color.dart';
import 'package:mapmyindia_gl/mapmyindia_gl.dart';

class CameraELocFeature extends StatefulWidget {
  @override
  State<StatefulWidget> createState() {
    return CameraELocFeatureState();
  }
}

class CameraELocFeatureState extends State {
  late MapmyIndiaMapController mapController;
  bool isMapLoded = false;

  static final CameraPosition _kInitialPosition = const CameraPosition(
    target: LatLng(25.321684, 82.987289),
    zoom: 14.0,
  );

  @override
  void initState() {
    super.initState();
    setState(() {
      isMapLoded = false;
    });
  }

  void _onMapCreated(MapmyIndiaMapController controller) {
    mapController = controller;
    setState(() {
      isMapLoded = true;
    });
  }

  @override
  Widget build(BuildContext context) {
    return Scaffold(
        appBar: AppBar(
          centerTitle: true,
          backgroundColor: MyColor.colorPrimary,
          brightness: Brightness.dark,
          title: Text(
            'Camera Feature',
            style: TextStyle(color: Colors.white),
          ),
          elevation: 0.2,
        ),
        body: Column(
          children: <Widget>[
            Expanded(
                child: MapmyIndiaMap(
              initialCameraPosition: _kInitialPosition,
              onMapCreated: _onMapCreated,
            )),
            isMapLoded
                ? Row(
                    mainAxisAlignment: MainAxisAlignment.start,
                    crossAxisAlignment: CrossAxisAlignment.center,
                    children: <Widget>[
                      SizedBox(
                        width: 4,
                      ),
                      Expanded(
                          child: FlatButton(
                        onPressed: () {
                          mapController.moveCameraWithELoc(
                              CameraELocUpdate.newELocBounds(['2T7S17', 'MMI000']));
                        },
                        child: Text(
                          'Move Camera',
                          style: TextStyle(color: Colors.white),
                        ),
                        color: MyColor.colorPrimaryDark,
                      )),
                      SizedBox(
                        width: 4,
                      ),
                      Expanded(
                          child: FlatButton(
                        onPressed: () {
                          mapController.easeCameraWithELoc(
                              CameraELocUpdate.newELocBounds(["5EU4EZ", "MMI000"]));
                        },
                        child: Text('Ease Camera',
                            style: TextStyle(color: Colors.white)),
                        color: MyColor.colorPrimaryDark,
                      )),
                      SizedBox(
                        width: 4,
                      ),
                      Expanded(
                          child: FlatButton(
                        onPressed: () {
                            mapController.animateCameraWithELoc(
                                CameraELocUpdate.newELocBounds(
                                    ['IB3BR9', 'MMI000']));
                        },
                        child: Text(
                          'Animate Camera',
                          style: TextStyle(color: Colors.white),
                          textAlign: TextAlign.center,
                        ),
                        color: MyColor.colorPrimaryDark,
                      )),
                      SizedBox(
                        width: 4,
                      )
                    ],
                  )
                : Container(),
          ],
        ));
  }
}
