import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:fluttertoast/fluttertoast.dart';
import 'package:mapmyindia_gl/mapmyindia_gl.dart';
import 'package:mapmyindia_gl_example/model/geoanlytics_model.dart';
import 'package:mapmyindia_gl_example/utils/color.dart';

class GeoAnalyticsWidget extends StatefulWidget {
  @override
  State<StatefulWidget> createState() {
    return GeoAnalyticsWidgetState();
  }
}

class GeoAnalyticsWidgetState extends State {
  static final CameraPosition _kInitialPosition = const CameraPosition(
    target: LatLng(25.321684, 82.987289),
    zoom: 10.0,
  );

  bool isStyleLoaded = false;
  // bool showBottomSheet = false;
  MapmyIndiaMapController? controller;
  List<GeoAnalyticsModel> requestModel = [
    GeoAnalyticsModel(
        type: MapmyIndiaGeoAnalyticsType.STATE,
        geoboundType: "stt_nme",
        // attribute: "t_p",
        propertyNames: ["stt_nme", "stt_id", "t_p"],
        requests: [
          MapmyIndiaGeoAnalyticsRequest(
              geoBound: "HARYANA",
              style: MapmyIndiageoAnalyticsAppearanceOption(
                  fillColor: '42a5f4',
                  fillOpacity: 0.5,
                  strokeColor: '000000',
                  labelSize: 10,
                  strokeWidth: 0,
                  labelColor: '000000')),
          MapmyIndiaGeoAnalyticsRequest(
              geoBound: "UTTAR PRADESH",
              style: MapmyIndiageoAnalyticsAppearanceOption(
                  fillColor: '42a5f4',
                  fillOpacity: 0.5,
                  strokeColor: '000000',
                  labelSize: 10,
                  strokeWidth: 0,
                  labelColor: '000000')),
          MapmyIndiaGeoAnalyticsRequest(
              geoBound: "ANDHRA PRADESH",
              style: MapmyIndiageoAnalyticsAppearanceOption(
                  fillColor: '42a5f4',
                  fillOpacity: 0.5,
                  strokeColor: '000000',
                  labelSize: 10,
                  strokeWidth: 0,
                  labelColor: '000000')),
          MapmyIndiaGeoAnalyticsRequest(
              geoBound: "KERALA",
              style: MapmyIndiageoAnalyticsAppearanceOption(
                  fillColor: '42a5f4',
                  fillOpacity: 0.5,
                  strokeColor: '000000',
                  labelSize: 10,
                  strokeWidth: 0,
                  labelColor: '000000')),
        ],
        geoBound: ["HARYANA", "UTTAR PRADESH", "ANDHRA PRADESH", "KERALA"]),
        //query: ">0"),
    GeoAnalyticsModel(
        type: MapmyIndiaGeoAnalyticsType.DISTRICT,
        geoboundType: "stt_nme",
        attribute: "t_p",
        propertyNames: ["dst_nme", "dst_id", "t_p"],
        requests: [
          MapmyIndiaGeoAnalyticsRequest(
              geoBound: "BIHAR",
              style: MapmyIndiageoAnalyticsAppearanceOption(
                  fillColor: '8bc34a',
                  fillOpacity: 0.5,
                  strokeColor: '000000',
                  labelSize: 10,
                  strokeWidth: 0,
                  labelColor: '000000')),
          MapmyIndiaGeoAnalyticsRequest(
              geoBound: "UTTARAKHAND",
              style: MapmyIndiageoAnalyticsAppearanceOption(
                  fillColor: '8bc34a',
                  fillOpacity: 0.5,
                  strokeColor: '000000',
                  labelSize: 10,
                  strokeWidth: 0,
                  labelColor: '000000'))
        ],
        geoBound: ["BIHAR", "UTTARAKHAND"],
        query: ">0"),
  ];

  @override
  Widget build(BuildContext context) {
    return Scaffold(
        appBar: AppBar(
          centerTitle: true,
          backgroundColor: MyColor.colorPrimary,
          systemOverlayStyle: SystemUiOverlayStyle.dark,
          title: Text(
            'GeoAnalytics Plugin',
            style: TextStyle(color: Colors.white),
          ),
          elevation: 0.2,
        ),
        body: Stack(
          children: [
            Column(children: [
              Expanded(
                  child: MapmyIndiaMap(
                    initialCameraPosition: _kInitialPosition,
                    // showGeoAnalyticsInfoWindow: false,
                    onMapError: (code, message) =>
                    {
                      print('$code -------- $message')
                    },
                    onGeoAnalyticsSelected: (list) => {
                      print("Geoanalytics Seleted: ${list.map((e) => e.toJson())}}")
                    },
                    onMapCreated: (map) =>
                    {
                      print("onMapCreated"),
                      controller = map
                    },
                    onStyleLoadedCallback: () =>
                    {
                      setState(() {
                        isStyleLoaded = true;
                      })
                    },
                  )),
              isStyleLoaded
                  ? TextButton(
                onPressed: () {
                  GeoAnalyticsModel model = requestModel[0];
                  model.isSelected = !model.isSelected;
                  hideOrShowgeoanalytics(model);
                  // showGeoAnalyticsList();
                },
                child: Text(
                  "Show GeoAnalytics List",
                  style: TextStyle(color: Colors.white),
                ),
                style: TextButton.styleFrom(
                    backgroundColor: MyColor.colorPrimaryDark,
                    textStyle: TextStyle(color: Colors.white)),
              )
                  : Container(),
            ]),
          ],
        )
    );
  }

  showGeoAnalyticsList() {
    showModalBottomSheet(
      context: context,
      enableDrag: true,
      builder: (context) {
        return Expanded(
            child: ListView.builder(
                itemCount: requestModel.length,
                itemBuilder: (context, index) {
                  return Container(
                      padding: EdgeInsets.fromLTRB(12, 0, 12, 0),
                      decoration: BoxDecoration(
                          color: Colors.white,
                          border: Border(bottom: BorderSide())),
                      child: Row(
                        children: [
                          Text(getGeoAnalyticsTypeName(
                              requestModel[index].type)),
                          Switch(
                              value: requestModel[index].isSelected,
                              onChanged: (isChecked) =>
                              {

                                setState(() {
                                  requestModel[index].isSelected = isChecked;
                                }),

                                if (isChecked)
                                  {
                                    moveCameraToSelectedGeoAnalytics(
                                        requestModel[index])
                                  },

                                hideOrShowgeoanalytics(
                                    requestModel[index]),
                              })
                        ],
                      ));
                }));
      },
    );
  }

  String getGeoAnalyticsTypeName(MapmyIndiaGeoAnalyticsType type) {
    switch (type) {
      case MapmyIndiaGeoAnalyticsType.STATE:
        return "state";
      case MapmyIndiaGeoAnalyticsType.DISTRICT:
        return "district";
      default:
        return "state";
    }
  }

  hideOrShowgeoanalytics(GeoAnalyticsModel requestModel) {
    if (requestModel.isSelected) {
      controller?.showGeoAnalytics(
          type: requestModel.type,
          geoboundType: requestModel.geoboundType,
          attribute: requestModel.attribute,
          geobound: requestModel.requests,
          query: requestModel.query,
          propertyNames: requestModel.propertyNames);
    } else {
      print("remove Geoanalytics");
      controller?.removeGeoAnalytics(type: requestModel.type,
          geoboundType: requestModel.geoboundType,
          attribute: requestModel.attribute,
          geobound: requestModel.requests,
          query: requestModel.query,
          propertyNames: requestModel.propertyNames);
    }
  }

  moveCameraToSelectedGeoAnalytics(GeoAnalyticsModel requestModel) async {
    try {
      GeoAnalyticsListResponse? geoAnalyticsResponse =
      await MapmyIndiaGeoAnalyticsList(
          getGeoAnalyticsTypeName(requestModel.type),
          requestModel.geoboundType,
          requestModel.geoBound,
          "b_box")
          .callGeoAnalyticsList();

      GeoAnalyticsListResult? results = geoAnalyticsResponse?.results;

      List<LatLng> latLngList = [];
      results?.getAttrValues?.forEach((element) {
        element.getAttrValues?.forEach((attr) {
          String bBox = attr['b_box'] as String;
          String truncateBox = bBox.substring(4, bBox.length - 1);
          String start = truncateBox.split(",")[0];
          String last = truncateBox.split(",")[1];
          latLngList.add(LatLng(double.parse(start.split(" ")[1]),
              double.parse(start.split(" ")[0])));
          latLngList.add(LatLng(double.parse(last.split(" ")[1]),
              double.parse(last.split(" ")[0])));
        });
      });
      if (latLngList.length > 0) {
        LatLngBounds bounds = boundsFromLatLngList(latLngList);
        controller?.animateCamera(CameraUpdate.newLatLngBounds(bounds,
            top: 10, left: 10, bottom: 10, right: 10));
      } else {
        print("No LatLng Bounds");
      }
    } catch (e) {
      if (e is PlatformException) {
        Fluttertoast.showToast(msg: '${e.code} --- ${e.message}');
      }
      print(e);
    }
  }

  boundsFromLatLngList(List<LatLng> list) {
    assert(list.isNotEmpty);
    double? x0, x1, y0, y1;
    for (LatLng latLng in list) {
      if (x0 == null || x1 == null || y0 == null || y1 == null) {
        x0 = x1 = latLng.latitude;
        y0 = y1 = latLng.longitude;
      } else {
        if (latLng.latitude > x1) x1 = latLng.latitude;
        if (latLng.latitude < x0) x0 = latLng.latitude;
        if (latLng.longitude > y1) y1 = latLng.longitude;
        if (latLng.longitude < y0) y0 = latLng.longitude;
      }
    }
    return LatLngBounds(
        northeast: LatLng(x1!, y1!), southwest: LatLng(x0!, y0!));
  }
}
