import 'dart:convert';
import 'dart:typed_data';

import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:mapmyindia_gl_example/utils/color.dart';
import 'package:mapmyindia_gl/mapmyindia_gl.dart';

class IndoorWidget extends StatefulWidget {
  @override
  State<StatefulWidget> createState() {
    return IndoorWidgetState();
  }
}

class IndoorWidgetState extends State {
  static final CameraPosition _kInitialPosition = const CameraPosition(
    target: LatLng(28.5425071, 77.1560724),
    zoom: 16.0,
  );

  late MapmyIndiaMapController controller;

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      appBar: AppBar(
        centerTitle: true,
        backgroundColor: MyColor.colorPrimary,
        systemOverlayStyle: SystemUiOverlayStyle.dark,
        title: Text(
          'Indoor',
          style: TextStyle(color: Colors.white),
        ),
        elevation: 0.2,
      ),
      body: MapmyIndiaMap(
        initialCameraPosition: _kInitialPosition,
        onMapError: (code, message) => {
          print('$code -------- $message')
        },
        onMapCreated: (map) =>
        {
          print("onMapCreated"),
          controller = map,
        },
        onStyleLoadedCallback: () => {addMarker()},
      ),
    );
  }

  /// Adds an asset image to the currently displayed style
  Future<void> addImageFromAsset(String name, String assetName) async {
    final ByteData bytes = await rootBundle.load(assetName);
    final Uint8List list = bytes.buffer.asUint8List();
    return controller.addImage(name, list);
  }

  void addMarker() async {
    await addImageFromAsset("icon", "assets/symbols/custom-icon.png");
    controller.addSymbol(SymbolOptions(
        geometry: LatLng(28.5425071, 77.1560724), iconImage: "icon"));
  }
}
